unit HuffUnit;

{$mode objfpc}{$H+}

////////////////////////////////////////////
//                                        //
//    Unit na vytvaranie a dekodovanie    //
//    Huffmanovej sifry                   //
//                                        //
//    Tomáš Vajda                         //
//    2.11. 2010                          //
//                                        //
////////////////////////////////////////////


interface

uses
  Classes, SysUtils, Dialogs;

type

  TAString = array[0..255] of string;

  THuffK = record
    Text: string;
    Kody: TAString;
  end;

  TPravdepodobnost = array[0..255] of Integer;

  { TVrchol }

  TVrchol = class
    Info: Char;
    Vyskyt: Integer;
    Left, Right: TVrchol;

    constructor Create(Z: Char; V: Integer; L: TVrchol = nil; R: TVrchol = nil);
    destructor Destroy; override;
  end;

  { TStrom }

  TStrom = class
  private
    Korene: array of TVrchol;
    Kody: TAString;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Spoj(V1, V2: TVrchol);
    procedure Pridaj(Z1: Char; V1: Integer; Z2: Char = #0; V2: Integer = 0);
    function Najmensi: TVrchol;
    function DruhyNajmensi: TVrchol;
    function VratINajmensieho: Integer;
    procedure OdstranKoren(K: TVrchol);
    procedure SpojNajmensieKorene;
    procedure Finalizacia;
    procedure VytvorHuffmanoveKody;
    procedure Backtracking(V: TVrchol; Kod: string);
    function Zakoduj(T: string): string;
    function VratKody: TAString;
  end;

  { TZnak }

  TZnak = class
  public
    Znak: Char;
    Vyskyt: Integer;
    Next: TZnak;

    constructor Create(Z: Char; V: Integer);
  end;

  { TZoznam }

  TZoznam = class
  private
    Polozky: TZnak;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Pridaj(Z: Char; V: Integer);
    procedure Odstran(Z: Char);
    function Vrat(I: Integer): TZnak;
    function Pocet: Integer;
  end;


function Huffman(TXT: string): THuffK;
function KodyText(K: TAString): string;
function DekodujHuffmana(TXT: string; K: TAString): string;
procedure Uloz(Zakodovane: string; Kody: TAString; NazovSuboru: string);
function Otvor(NazovSuboru: string): THuffK;

implementation

function VyratajPravdepodobnost(TXT: string): TPravdepodobnost;
var
  I: Integer;
begin
  for I := 0 to 255 do
    Result[I] := 0;

  for I := 1 to Length(TXT) do
    inc(Result[Ord(TXT[I])]);
end;

function Huffman(TXT: string): THuffK;
var
  Zoznam: TZoznam;
  Strom: TStrom;
  I: Integer;
  Z1, Z2: Char;
  V1, V2: Integer;
  PP: TPravdepodobnost;
begin
  PP := VyratajPravdepodobnost(TXT);

  Zoznam := TZoznam.Create;

  for I := 0 to 255 do
    if PP[I] > 0 then
      Zoznam.Pridaj(Chr(I), PP[I]);


  Strom := TStrom.Create;

  while Zoznam.Pocet > 0 do
  begin
    Z1 := Zoznam.Vrat(1).Znak;
    V1 := Zoznam.Vrat(1).Vyskyt;

    if Strom.DruhyNajmensi <> nil then
    begin
      if V1 <= Strom.DruhyNajmensi.Vyskyt then
        Zoznam.Odstran(Zoznam.Vrat(1).Znak)
      else
      begin
        Strom.SpojNajmensieKorene;
        Continue;
      end;
    end
    else
      Zoznam.Odstran(Zoznam.Vrat(1).Znak);

    if Zoznam.Vrat(1) <> nil then
    begin
      Z2 := Zoznam.Vrat(1).Znak;
      V2 := Zoznam.Vrat(1).Vyskyt;
    end;

    if ((Zoznam.Vrat(1) <> nil) and (Strom.Najmensi = nil))
    or (Zoznam.Vrat(1) <> nil) and (Strom.Najmensi.Vyskyt > V2) then
    begin
      Strom.Pridaj(Z1, V1, Z2, V2);
      Zoznam.Odstran(Zoznam.Vrat(1).Znak);
    end
    else
    begin
      Strom.Pridaj(Z1, V1);
    end;
  end;

  Strom.Finalizacia;
  Strom.VytvorHuffmanoveKody;

  Result.Text := Strom.Zakoduj(TXT);
  Result.Kody := Strom.VratKody;

  Zoznam.Free;
  Strom.Free;
end;

function KodyText(K: TAString): string;
var
  I: Integer;
begin
  Result := '';

  for I := 0 to High(K) do
    if K[I] <> '' then
      Result := Result + #39 + Chr(I) + #39 +  ': ' + K[I] + #13#10;

  if Result <> '' then
    SetLength(Result, Length(Result) - 2);
end;

function DekodujHuffmana(TXT: string; K: TAString): string;
var
  I, J: Integer;
  Kod: string;
begin
  Result := '';
  Kod := '';

  for I := 1 to Length(TXT) do
  begin
    Kod := Kod + TXT[I];

    for J := 0 to High(K) do
      if K[J] = Kod then
      begin
        Result := Result + Chr(J);
        Kod := '';
        Break;
      end;
  end;
end;

function PocetKodov(Kody: TAString): Integer;
var
  B: Byte;
begin
  Result := 0;

  for B := 0 to High(Kody) do
    if Kody[B] <> '' then
      inc(Result);
end;

function VratPocetNul(I: Integer): string;
begin
  Result := '';

  while I > 0 do
  begin
    Result := Result + '0';
    dec(I);
  end;
end;

function Mocnina(C, E: Byte): Byte;
begin
  if E = 0 then
  begin
    Result := 1;
    Exit;
  end;

  Result := C * Mocnina(C, E - 1);
end;

function BinStrToByte(BinStr: string): Byte;
var
  I: Integer;
begin
  Result := 0;

  for I := 1 to 8 do
    if BinStr[I] = '1' then
      Result := Result + Mocnina(2, 8 - I);
end;

function ByteToBinStr(B: Byte): string;
begin
  Result := '';

  while B > 0 do
  begin
    if B mod 2 = 1 then
      Result := '1' + Result
    else
      Result := '0' + Result;

    B := B div 2;
  end;

  if Length(Result) < 8 then
    Result := VratPocetNul(8 - Length(Result)) + Result;
end;

procedure Uloz(Zakodovane: string; Kody: TAString; NazovSuboru: string);
var
  B, L: Byte;
  I, J: Integer;
  B8: string;
  Subor: TStream;
begin
  Subor := TFileStream.Create(NazovSuboru, fmCreate);


  // ulozenie kodov
  I := PocetKodov(Kody);
  Subor.Write(I, SizeOf(Integer));

  for B := 0 to High(Kody) do
    if Kody[B] <> '' then
    begin
      Subor.Write(B, SizeOf(Byte));
      L := Length(Kody[B]);
      Subor.Write(L, SizeOf(Byte));
      Subor.Write(Kody[B, 1], L);
    end;


  // ulozenie samotneho komprimovaneho textu
  I := Length(Zakodovane);
  Subor.Write(I, SizeOf(Integer));

  for J := 0 to (I - 1) div 8 do
  begin
    B8 := Copy(Zakodovane, J * 8 + 1, 8);

    if Length(B8) < 8 then
      B8 := B8 + VratPocetNul(8 - Length(B8));

    B := BinStrToByte(B8);
    Subor.Write(B, SizeOf(Byte));
  end;

  Subor.Free;
end;

function Otvor(NazovSuboru: string): THuffK;
var
  I, J: Integer;
  B, L: Byte;
  Subor: TStream;
begin
  Subor := TFileStream.Create(NazovSuboru, fmOpenRead);

  for I := 0 to High(Result.Kody) do
    Result.Kody[I] := '';


  // nacitanie kodov
  Subor.Read(I, SizeOf(Integer));

  for J := 1 to I do
  begin
    Subor.Read(B, SizeOf(Byte));
    Subor.Read(L, SizeOf(Byte));
    SetLength(Result.Kody[B], L);
    Subor.Read(Result.Kody[B, 1], L);
  end;


  // nacitanie zakodovaneho textu
  Subor.Read(I, SizeOf(Integer));
  Result.Text := '';

  while Subor.Position < Subor.Size do
  begin
    Subor.Read(B, SizeOf(Byte));
    Result.Text := Result.Text + ByteToBinStr(B);
  end;

  SetLength(Result.Text, I);

  Subor.Free;
end;

{ TZnak }

constructor TZnak.Create(Z: Char; V: Integer);
begin
  Znak := Z;
  Vyskyt := V;
  Next := nil;
end;

{ TZoznam }

constructor TZoznam.Create;
begin
  while Pocet > 0 do
    Odstran(Vrat(1).Znak);
end;

destructor TZoznam.Destroy;
var
  Z, ZP: TZnak;
begin
  while Pocet > 0 do
  begin
    ZP := nil;
    Z := Polozky;

    while Z.Next <> nil do
    begin
      ZP := Z;
      Z := Z.Next;
    end;

    Z.Free;

    if ZP <> nil then
      ZP.Next := nil;
  end;
end;

procedure TZoznam.Pridaj(Z: Char; V: Integer);
var
  ZPP, ZP, Temp: TZnak;
begin
  if Polozky = nil then
  begin
    Polozky := TZnak.Create(Z, V);
  end
  else if Polozky.Next = nil then
  begin
    Temp := TZnak.Create(Z, V);

    if V > Polozky.Vyskyt then
      Polozky.Next := Temp
    else
    begin
      Temp.Next := Polozky;
      Polozky := Temp;
    end;
  end
  else
  begin
    ZPP := nil;
    ZP := Polozky;

    while (ZP <> nil) and (V > ZP.Vyskyt) do
    begin
      ZPP := ZP;
      ZP := ZP.Next;
    end;

    Temp := TZnak.Create(Z, V);
    if ZP = nil then
    begin
      ZPP.Next := Temp;
    end
    else
    begin
      if ZPP <> nil then
      begin
        Temp.Next := ZPP.Next;
        ZPP.Next := Temp;
      end
      else
      begin
        Temp.Next := Polozky;
        Polozky := Temp;
      end;
    end;
  end;
end;

procedure TZoznam.Odstran(Z: Char);
var
  ZP, Temp: TZnak;
begin
  ZP := Polozky;

  while (ZP <> nil) and (ZP.Znak <> Z) do
    ZP := ZP.Next;

  if ZP <> nil then
  begin
    if Polozky <> ZP then
    begin
      ZP.Free;
    end
    else
    begin
      if Polozky.Next <> nil then
      begin
        Temp := Polozky;
        Polozky := Polozky.Next;
        Temp.Free;
      end
      else
      begin
        Polozky.Free;
        Polozky := nil;
      end;
    end;
  end;
end;

function TZoznam.Vrat(I: Integer): TZnak;
var
  Z: TZnak;
  J: Integer;
begin
  Z := Polozky;
  J := 0;

  while Z <> nil do
  begin
    inc(J);

    if I = J then
      break;

    Z := Z.Next;
  end;

  Result := Z;
end;

function TZoznam.Pocet: Integer;
var
  Z: TZnak;
  I: Integer;
begin
  Z := Polozky;
  I := 0;

  while Z <> nil do
  begin
    inc(I);
    Z := Z.Next;
  end;

  Result := I;
end;

{ TVrchol }

constructor TVrchol.Create(Z: Char; V: Integer; L: TVrchol; R: TVrchol);
begin
  Info := Z;
  Left := L;
  Right := R;
  Vyskyt := V;
end;

destructor TVrchol.Destroy;
begin
  if Left <> nil then
    Left.Free;

  if Right <> nil then
    Right.Free;
end;

{ TStrom }

constructor TStrom.Create;
var
  I: Integer;
begin
  SetLength(Korene, 0);

  for I := 0 to 255 do
    Kody[I] := '';
end;

destructor TStrom.Destroy;
var
  I: Integer;
begin
  for I := 0 to High(Korene) do
    Korene[I].Free;
end;

procedure TStrom.Spoj(V1, V2: TVrchol);
begin
  SetLength(Korene, Length(Korene) + 1);

  if V1.Vyskyt < V2.Vyskyt then
    Korene[High(Korene)] := TVrchol.Create(#0, V1.Vyskyt + V2.Vyskyt, V1, V2)
  else
    Korene[High(Korene)] := TVrchol.Create(#0, V1.Vyskyt + V2.Vyskyt, V2, V1);
end;

procedure TStrom.Pridaj(Z1: Char; V1: Integer; Z2: Char; V2: Integer);
var
  Vrch1, Vrch2: TVrchol;
begin
  if Z2 = #0 then
  begin
    Vrch1 := TVrchol.Create(Z1, V1);
    Vrch2 := Najmensi;

    if Vrch1.Vyskyt < Vrch2.Vyskyt then
      Korene[VratINajmensieho] := TVrchol.Create(#0, V1 + Vrch2.Vyskyt, Vrch1, Vrch2)
    else
      Korene[VratINajmensieho] := TVrchol.Create(#0, V1 + Vrch2.Vyskyt, Vrch2, Vrch1);
  end
  else
  begin
    SetLength(Korene, Length(Korene) + 1);

    if V1 < V2 then
    begin
      Vrch1 := TVrchol.Create(Z1, V1);
      Vrch2 := TVrchol.Create(Z2, V2);
    end
    else
    begin
      Vrch2 := TVrchol.Create(Z1, V1);
      Vrch1 := TVrchol.Create(Z2, V2);
    end;

    Korene[High(Korene)] := TVrchol.Create(#0, V1 + V2, Vrch1, Vrch2);
  end;
end;

function TStrom.Najmensi: TVrchol;
var
  I, Naj: Integer;
begin
  Result := nil;
  Naj := MAXINT;

  for I := 0 to High(Korene) do
    if Korene[I].Vyskyt < Naj then
    begin
      Naj := Korene[I].Vyskyt;
      Result := Korene[I];
    end;
end;

function TStrom.DruhyNajmensi: TVrchol;
var
  I, Naj, Naj2: Integer;
begin
  Result := nil;
  Naj := MAXINT;

  for I := 0 to High(Korene) do
    if Korene[I].Vyskyt < Naj then
      Naj := Korene[I].Vyskyt;

  Naj2 := MAXINT;

  for I := 0 to High(Korene) do
    if (Korene[I].Vyskyt <> Naj) and (Korene[I].Vyskyt < Naj2) then
    begin
      Naj2 := Korene[I].Vyskyt;
      Result := Korene[I];
    end;
end;

function TStrom.VratINajmensieho: Integer;
var
  I, Naj: Integer;
begin
  Result := -1;
  Naj := MAXINT;

  for I := 0 to High(Korene) do
    if Korene[I].Vyskyt < Naj then
    begin
      Naj := Korene[I].Vyskyt;
      Result := I;
    end;
end;

procedure TStrom.OdstranKoren(K: TVrchol);
var
  NKorene: array of TVrchol;
  I: Integer;
begin
  SetLength(NKorene, 0);

  for I := 0 to High(Korene) do
  begin
    if Korene[I] <> K then
    begin
      SetLength(NKorene, Length(NKorene) + 1);
      NKorene[High(NKorene)] := Korene[I];
    end;
  end;

  SetLength(Korene, Length(NKorene));
  Korene := NKorene;
end;

procedure TStrom.SpojNajmensieKorene;
var
  K1, K2: TVrchol;
begin
  K1 := Najmensi;
  OdstranKoren(K1);
  K2 := Najmensi;
  OdstranKoren(K2);

  Spoj(K1, K2);
end;

procedure TStrom.Finalizacia;
begin
  while Length(Korene) > 1 do
    SpojNajmensieKorene;
end;

procedure TStrom.VytvorHuffmanoveKody;
begin
  if Length(Korene) = 1 then
    Backtracking(Korene[0], '');
end;

procedure TStrom.Backtracking(V: TVrchol; Kod: string);
begin
  if V.Left <> nil then
  begin
    Backtracking(V.Left, Kod + '0');
  end;

  if V.Right <> nil then
  begin
    Backtracking(V.Right, Kod + '1');
  end;

  if (V.Left = nil) and (V.Right = nil) then
  begin
    Kody[Ord(V.Info)] := Kod;
  end;
end;

function TStrom.Zakoduj(T: string): string;
var
  I: Integer;
begin
  Result := '';

  for I := 1 to Length(T) do
    Result := Result + Kody[Ord(T[I])];
end;

function TStrom.VratKody: TAString;
begin
  Result := Kody;
end;

end.

